#!/usr/bin/perl -w
#
# index.cgi
# John Simpson <jms1@jms1.net>
#
# Verify authorization, then show email processing rules relevant to the
# currently focused entity (mailbox, domain, or the entire system.)
#
# $Id: index.cgi 114 2012-07-22 01:22:46Z jms1 $
#
###############################################################################
#
# Copyright (C) 2012 John M. Simpson.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 or version 3 of the
# license, at your option.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

require 5.003 ;
use strict ;
use CGI qw ( :standard ) ;

use Rules ;

my ( @phases , %phaselevel , %phasename ) ;

###############################################################################

sub nextphase($)
{
	my $cur = shift ;

	if ( $cur < 0 )
	{
		return $phases[0] ;
	}

	for my $n ( 0 .. $#phases )
	{
		if ( $cur == $phases[$n] )
		{
			return ( $phases[$n+1] || undef ) ;
		}
	}
	return undef ;
}

###############################################################################
#
# convert seconds to hours:minutes:seconds

sub hms($)
{
	my $s = shift ;

	my $h = int ( $s / 3600 ) ;
	$s -= 3600 * $h ;

	my $m = int ( $s / 60 ) ;
	$s -= 60 * $m ;

	if ( $h > 0 )
	{
		return sprintf ( '%d:%02d:%02d' , $h , $m , $s ) ;
	}

	if ( $m > 0 )
	{
		return sprintf ( '%d:%02d' , $m , $s ) ;
	}

	return sprintf ( '0:%02d' , $s ) ;
}

###############################################################################
###############################################################################
###############################################################################

require_login() ;

########################################
# Read phase info into memory.

my $query = "SELECT phase , level , description FROM phases ORDER BY phase" ;
my @data = SQL_select ( $query ) ;
while ( $#data > -1 )
{
	my $p = shift @data ;
	my $l = shift @data ;
	my $n = shift @data ;

	push ( @phases , $p ) ;
	$phaselevel{$p} = $l ;
	$phasename{$p} = $n ;
}

########################################
# Can't use the system with no phases.

if ( $#phases < 0 )
{
	fatal 'No phases have been configured.' ;
}

########################################
# Figure the correct "WHERE" clauses for the focus we're looking at.
# $focus will be one of the following:
#	@		System-wide rules
#	@domain		Domain-wide rules
#	blah@domain	Mailbox rules

my $rwhere = SQL_quote ( $focus ) . ' LIKE recipient' ;

my $lwhere = 'recipient LIKE ' . SQL_quote ( $focus ) ;
if ( $focus eq '@' )
{
	$lwhere = SQL_true() ;
}
elsif ( $focus =~ m|^\@| )
{
	$lwhere = "recipient LIKE '%$focus'" ;
}

########################################
# Fake the old "mode" cookie (access mode)

my $mode = 1 ;
if ( $focus eq "\@" )
{
	$mode = 3 ;
}
elsif ( $focus =~ m|^\@| )
{
	$mode = 2 ;
}

########################################
# Count how many rules are in each phase.

$query = "SELECT phase , COUNT(*) FROM rules WHERE $rwhere"
	. ' GROUP BY phase ORDER BY phase' ;
my %phase_count = SQL_select ( $query ) ;

########################################
# Build the query and grab the data.

$query = "SELECT id , phase , seq , recipient , accept , type , sender ,"
	. " ip , rbl , delay , target , description ,"
	. " ( SELECT COUNT(*) FROM log WHERE $lwhere"
	. " AND log.ruleid = rules.id )"
	. " FROM rules WHERE $rwhere ORDER BY phase , seq" ;

@data = SQL_select ( $query ) ;

########################################
# Process the data.

my $list	= '' ;
my $lphase	= -1 ;
my $rcount	= 0 ;
my $seen_final	= 0 ;
my $after_final	= 0 ;

while ( $#data > -1 )
{
	########################################
	# read rule from the database query

	my $id		=   shift @data ;
	my $phase	=   shift @data ;
	my $seq		=   shift @data ;
	my $recipient	=   shift @data ;
	my $accept	=   shift @data ;
	my $type	=   shift @data ;
	my $sender	= ( shift @data || '' ) ;
	my $ip		= ( shift @data || '' ) ;
	my $rbl		= ( shift @data || '' ) ;
	my $delay	= ( shift @data || 0  ) ;
	my $target	= ( shift @data || '' ) ;
	my $description	= ( shift @data || '' ) ;
	my $matches	= ( shift @data || 0  ) ;

	########################################
	# figure out how to display it

	my $zdisp = ( $accept ? 'ACCEPT' : 'REJECT' ) ;
	my $ztype = ( $typename{$type} || "($type)" ) ;

	my $zdata = '(???)' ;

	   if ( $type eq 'A' ) { $zdata = '(all)'		; }
	elsif ( $type eq 'E' ) { $zdata = "<tt>$sender</tt>"	; }
	elsif ( $type eq 'I' ) { $zdata = "<tt>$ip</tt>"	; }
	elsif ( $type eq 'R' ) { $zdata = "<i>4.3.2.1.</i><tt>$rbl</tt>" ; }
	elsif ( $type eq 'T' ) { $zdata = "<tt>$target</tt>"	; }
	elsif ( $type eq 'D' ) { $zdata = '' ; $zdisp = 'DEBUG' ; }
	elsif ( $type eq 'U' ) { $zdata = '(Authenticated clients)' ; }
	elsif ( $type eq 'G' )
	{
		$zdata = $delay . ' (' . hms ( $delay ) . ')' ;
		$zdisp = 'DELAY' ;
	}

	my $zbg = ( ( ( $rcount % 6 ) > 2 ) ? "rule_bg_a" : "rule_bg_b" ) ;

	my $zcmds = "" ;
	if ( $phaselevel{$phase} == $mode )
	{
		########################################
		# command links for the current rule

		$zcmds .= "<a href='add.cgi?phase=$phase&seq=$seq'>"
			. "<img src='button-add-below.png' alt='[ADD]'"
			. " border='0'></a>" ;
		$zcmds .= "<a href='del.cgi?phase=$phase&seq=$seq'>"
			. "<img src='button-delete.png' alt='[DEL]'"
			. " border='0'></a>" ;
		if ( rule_can_edit ( $type ) )
		{
			$zcmds .= "<a href='edit.cgi?phase=$phase&seq=$seq'>"
				. "<img src='button-edit.png' alt='[EDIT]'"
				. " border='0'></a>" ;
		}
		else
		{
			$zcmds .= "<img src='fill-button.png' alt=''"
				. " border='0'>" ;
		}

		if ( $seq > 1 )
		{
			my $prev = $seq - 1 ;
			$zcmds .= "<a href='swap.cgi?phase=$phase&seq=$prev'>"
				. "<img src='button-move-up.png' alt='[Up]'"
				. " border='0'></a>" ;
		}
		else
		{
			$zcmds .= "<img src='fill-button.png' alt=''"
				. " border='0'>" ;
		}

		if ( $seq < $phase_count{$phase} )
		{
			$zcmds .= "<a href='swap.cgi?phase=$phase&seq=$seq'>"
				. "<img src='button-move-down.png' alt='[Down]'"
				. " border='0'></a>" ;
		}
		else
		{
			$zcmds .= "<img src='fill-button.png' alt=''"
				. " border='0'>" ;
		}


	}

	########################################
	# if this rule's phase is not the same as the previous rule,
	# we need phase header lines before this rule.

	while ( $lphase && ( $lphase < $phase ) )
	{
		$lphase = nextphase ( $lphase ) ;

		my $pcmd = '' ;
		my $zclass = 'phaseline' ;

		if ( $phaselevel{$lphase} == $mode )
		{
			$zclass = 'currentphaseline' ;

			########################################
			# phase commands

			$pcmd .= "<a href='add.cgi?phase=$lphase&seq=0'>"
				. "<img src='button-add-below.png' alt='[ADD]'"
				. " border='0'></a>" ;
		}

		$list .= <<EOF ;
  <tr valign='middle'>
    <td class='$zclass' colspan='5'>[$lphase] $phasename{$lphase}</td>
    <td class='$zclass sep'>$pcmd</td>
    <td class='$zclass sep'></td>
  </tr>
EOF
	}
	$lphase = $phase ;

	########################################
	# add this rule to the output

	my $fstart = $seen_final ? "<span class='will_not_run'>" : '' ;
	my $fend   = $seen_final ? '</span>' : '' ;

	my $zmatch = comma ( $matches ) ;

	$list .= <<EOF ;
  <tr valign="middle">
    <!-- id="$id" recipient="$recipient" -->
    <td class="rulecell $zbg" align="right">$fstart$seq$fend</td>
    <td class="rulecell sep $zbg" align="center">$fstart$zdisp$fend</td>
    <td class="rulecell sep $zbg" align="center">$fstart$ztype$fend</td>
    <td class="rulecell sep $zbg" align="left">$fstart$zdata$fend</td>
    <td class="rulecell sep $zbg" align="right">$zmatch</td>
    <td class="rulecell sep $zbg nobr" align="left">$zcmds</td>
    <td class="rulecell sep $zbg" align="left">$description</td>
  </tr>
EOF

	########################################
	# If this was a "final" rule, remember it so we can mark the
	# remaining rules as "will not be processed".

	if ( $seen_final )
	{
		$after_final ++ ;
	}

	if ( rule_is_final ( $type ) )
	{
		$seen_final = 1 ;
	}

	########################################
	# move on to the next rule

	$rcount ++ ;
}

########################################
# finish list of phase header lines

$lphase = nextphase ( $lphase ) ;
while ( $lphase )
{
	my $zclass = 'phaseline' ;
	my $pcmd   = '' ;

	if ( $phaselevel{$lphase} == $mode )
	{
		$zclass = 'currentphaseline' ;

		########################################
		# phase commands

		$pcmd .= "<a href='add.cgi?phase=$lphase&seq=0'>"
			. "<img src='button-add-below.png' alt='[ADD]'"
			. " border='0'></a>" ;
	}

	$list .= <<EOF ;
  <tr valign='middle'>
    <td class='$zclass' colspan='5'>[$lphase] $phasename{$lphase}</td>
    <td class='$zclass sep'>$pcmd</td>
    <td class='$zclass sep'></td>
  </tr>
EOF

	$lphase = nextphase ( $lphase ) ;
}

########################################
# If we had to show any "will not run" rules, show an explanation

my $note = '' ;
if ( $after_final )
{
	$note = <<EOF ;

<div class='notebox'>Note: the rules in <span class='will_not_run'>smaller
italic type</span> at the end of the list will not be executed by the engine
for this recipient, because they come after a "final" rule (either an
all-messages rule, or a greylisting rule.)</div>

EOF
}

########################################
# header for list of rules

my $list_header = "Rules affecting email sent to $focus" ;

if ( $focus eq '@' )
{
	$list_header = 'Rules affecting all mail sent to the server' ;
}
elsif ( $focus =~ m|^\@(.+)| )
{
	my $domain = $1 ;
	$list_header = "Rules affecting all mail sent to domain \@$domain" ;
}

########################################
# get list of mailboxes which have rules

my @mailboxes = () ;

if ( $access > 1 )	# domain admin or system admin
{
	if ( $focus =~ m|\@(.+)| )
	{
		my $domain = $1 ;

		$query = 'SELECT DISTINCT recipient FROM rules'
			. ' WHERE recipient LIKE '
			. SQL_quote ( "\%\@$domain" )
			. ' ORDER BY recipient' ;
		my @rcpt = SQL_select ( $query ) ;

		for my $r ( @rcpt )
		{
			next if ( $r =~ m|^\%\@| ) ;
			push ( @mailboxes , $r ) ;
		}
	}
}

########################################
# focus list - list of entities who own at least one rule

my $focus_list = '' ;

if ( $ENV{'SHOW_FOCUS_LIST'} )
{
	if ( $access > 2 )
	{
		$focus_list = get_focus_list() ;
	}
	elsif ( $access > 1 )
	{
		( my $dom = $userid ) =~ s|^.*\@|| ;
		$focus_list = get_focus_list ( $dom ) ;
	}
}

########################################
# focus form shows up after the main rules list

my $focus_form = '' ;

if ( $access > 2 )	# system admin
{
	my $ssel = '' ;
	my $dsel = '' ;
	my $dval = '' ;
	my $msel = '' ;
	my $mval = '' ;
	my $xsel = '' ;

	if ( $focus eq '@' )
	{
		$ssel = 'checked' ;
	}
	elsif ( $focus =~ m|^\@(.+)| )
	{
		$dval = $1 ;
		$dsel = 'checked' ;
	}
	elsif ( $focus && ( $focus ne $userid ) )
	{
		$mval = $focus ;
		$msel = 'checked' ;
	}
	else
	{
		$xsel = 'checked' ;
	}

	$focus_form = parse_template ( 'focus-sa.txt',
		'ssel' => $ssel ,
		'dsel' => $dsel ,
		'dval' => $dval ,
		'msel' => $msel ,
		'mval' => $mval ,
		'xsel' => $xsel ,
		'focuslist' => $focus_list ) ;
}
elsif ( $access > 1 )	# domain admin
{
	my $dsel = '' ;
	my $usel = '' ;
	my $uval = '' ;
	my $xsel = '' ;

	( my $dom = $userid ) =~ s|^.*\@|| ;

	if ( $focus =~ m|^\@(.+)| )
	{
		$dsel = 'checked' ;
	}
	elsif ( $focus && ( $focus ne $userid ) )
	{
		$usel = 'checked' ;
		$uval = $focus ;
	}
	else
	{
		$xsel = 'checked' ;
	}

	$focus_form = parse_template ( 'focus-da.txt' ,
		'dsel' => $dsel ,
		'dom'  => $dom  ,
		'usel' => $usel ,
		'uval' => $uval ,
		'xsel' => $xsel ,
		'focuslist' => $focus_list ) ;
}

########################################
# show the list already

print page_start ( "Main Menu" ) ;
print parse_template ( "main-menu.txt" ,
	'list_header'	=> $list_header ,
	'list'		=> $list ,
	'note'		=> $note ,
	'focus_form'	=> $focus_form ) ;
print page_end() ;
exit 0 ;
