# Rules.pm
# John Simpson <jms1@jms1.net> 2008-10-26
#
# Library of functions used by the qmail rules web interface
#
# $Id: Rules.pm 129 2012-07-24 04:30:58Z jms1 $
#
###############################################################################
#
# Copyright (C) 2012 John M. Simpson.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 or version 3 of the
# license, at your option.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

package Rules ;
require 5.003 ;
use strict ;
use Carp ;
use CGI qw ( :standard ) ;
use DBI ;
use Digest::MD5 qw ( md5_base64 ) ;
use MIME::Base64 ;
use Socket ;

###############################################################################
#
# startup stuff

use vars qw
(
	$dbconn $dbuser $dbpass $dbh $sql_log
	$sitename $approot $css_url
	$docroot $site_url $app_url $request_uri $remote_ip $web_proto
	$ses_timeout $rc_session $rc_focus
	$userid $access $login_err $session $focus %typename $g4ver
) ;

my ( $fnsth ) ;

BEGIN
{
	use Exporter ;
	use vars qw ( $VERSION @ISA @EXPORT @EXPORT_OK ) ;

	$VERSION	= 1.0 ;
	@ISA		= qw ( Exporter ) ;
	@EXPORT		= qw ( logit bad fatal
				SQL_connect SQL_ping SQL_quote SQL_do
				SQL_select SQL_selectfirst SQL_selectnext
				SQL_transaction SQL_commit SQL_rollback
				SQL_true SQL_false

				comma
				read_file read_template parse_string
				parse_file parse_template

				validate_email validate_ip validate_rbl
				validate_grey validate_regex

				redirect_url redirect_self redirect_home

				get_session create_session end_session
				set_focus focus_string get_focus_list
				rule_is_final rule_can_edit

				init 

				page_start page_end

				check_vpopmaild require_login

				$dbconn $dbuser $dbpass
				$sitename $approot $css_url
				$site_url $app_url $docroot $request_uri
				$remote_ip $web_proto $ses_timeout

				$userid $access $login_err $session $focus
				%typename
			) ;
	@EXPORT_OK	= qw ( ) ;

	$dbh		= undef ;

	$g4ver		= '(???)' ;
	$sql_log	= ( ( ( $ENV{'G4SQLLOG'} || '' ) eq '1' ) ? 1 : 0 ) ;
	$sitename	= ( $ENV{'SITE_NAME'} || "Qmail Rules Interface" ) ;
	$ses_timeout	= ( $ENV{'SESSION_TIMEOUT'} || 1800 ) ;

	%typename =
	(
		'A' => 'All Messages'	,
		'D' => 'Debug'		,
		'E' => 'From Sender'	,
		'G' => 'Greylist' 	,
		'I' => 'From IP'	,
		'R' => 'RBL match'	,
		'T' => 'To Recipient'	,
		'U' => 'AUTH client'	,
	) ;
}

END
{
	if ( $dbh )
	{
		$dbh->disconnect() ;
		$dbh = undef ;
	}
}

###############################################################################
#
# Send a message to the apache error_log file.

sub logit($)
{
	print STDERR $_[0] ;
}

###############################################################################
#
# Generic error handler for bad input.

sub bad($)
{
	logit "bad: $_[0]" ;

	print <<EOF ;
Content-type: text/html

<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01//EN"
   "http://www.w3.org/TR/html4/strict.dtd">
<html>
<head>
<title>$sitename - Try again.</title>
<link rel="stylesheet" type="text/css" href="$css_url">
</head>
<body>
<h1>Try again.</h1>

<p>The following error condition has been detected, probably caused by
invalid input:</p>

<p class="codeblock">$_[0]</p>

<p>Please use your browser's BACK button, correct the mistake, and try
again.</p>

</body>
</html>
EOF

	exit 0 ;
}

###############################################################################
#
# Generic error handler for fatal errors.

sub fatal($)
{
	logit "fatal: $_[0]" ;

	print <<EOF ;
Status: 500 Internal server error
Content-type: text/html

<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01//EN"
   "http://www.w3.org/TR/html4/strict.dtd">
<html>
<head>
<title>$sitename - Internal Error</title>
<link rel="stylesheet" type="text/css" href="$css_url">
</head>
<body>
<h1>Internal Error</h1>

<p>The following error condition has been detected, probably caused by
something beyond your control.</p>

<p class="codeblock">$_[0]</p>

<p>It's doubtful you will be able to correct this problem. Please don't try
again, as this will only cause multiple error loggings and may cause your IP
address to be automatically banned from the system.</p>

</body>
</html>
EOF

	exit 0 ;
}

###############################################################################
#
# connect to the database

sub SQL_connect()
{
	unless ( $dbh )
	{
		die ( "ERROR: no database connection string configured\n" )
			unless ( $dbconn ) ;

		$dbh ||= DBI->connect ( $dbconn , $dbuser , $dbpass ) ;

		unless ( $dbh )
		{
			my $err = $DBI::errstr ;
			die ( "Cannot connect to \"$dbconn\": $err\n" ) ;
		}
	}
}

###############################################################################
#
# Make sure the SQL connection works, by doing a (non-logged) "SELECT 1" query.
# Returns 1 if OK, 0 if not.

sub SQL_ping()
{
	eval { SQL_connect() ; } ;
	if ( $@ )
	{
		logit "$@\n" ;
		return 0 ;
	}
	$dbh || return 0 ;

	my $olog = $sql_log ;
	$sql_log = 0 ;

	my ( $ok ) = SQL_select ( 'SELECT 1' ) ;

	$sql_log = $olog ;

	return ( $ok ? 1 : 0 ) ;
}

###############################################################################
#
# do SQL quoting

sub SQL_quote($)
{
	my $data = shift ;

	SQL_connect() ;

	return $dbh->quote ( $data ) ;
}

###############################################################################
#
# handle a "do" query (insert, delete, update)

sub SQL_do($)
{
	my $query = shift ;

	$sql_log && print STDERR "SQL: $query\n" ;

	SQL_connect() ;

	$dbh->do ( $query )
		or die ( "do($query): " . $dbh->errstr . "\n" ) ;
}

###############################################################################
#
# handle a select query

sub SQL_select($)
{
	my $query = shift ;

	$sql_log && print STDERR "SQL: $query\n" ;

	SQL_connect() ;

	my $sth = $dbh->prepare ( $query )
		or die ( "prepare($query): " . $dbh->errstr . "\n" ) ;

	$sth->execute()
		or die ( "execute($query): " . $dbh->errstr . "\n" ) ;

	my @rv = () ;

	while ( my @data = $sth->fetchrow_array() )
	{
		push ( @rv , @data ) ;
	}

	$sth->finish() ;
	return wantarray ? @rv : ( $#rv > -1 ? $rv[0] : undef ) ;
}

###############################################################################
#
# selectfirst and selectnext functions

sub SQL_selectfirst($)
{
	my $query = shift ;

	$sql_log && print STDERR "SQL: $query\n" ;

	SQL_connect() ;

	if ( $fnsth )
	{
		$fnsth->finish() ;
		undef $fnsth ;
	}

	$fnsth = $dbh->prepare ( $query )
		or die ( "prepare($query): " . $dbh->errstr . "\n" ) ;

	$fnsth->execute()
		or die ( "execute($query): " . $dbh->errstr . "\n" ) ;

	my @rv = $fnsth->fetchrow_array() ;

	unless ( @rv )
	{
		$fnsth->finish() ;
		$fnsth = undef ;
	}

	return @rv ;
}

sub SQL_selectnext()
{
	my @rv = undef ;

	if ( $fnsth )
	{
		@rv = $fnsth->fetchrow_array() ;

		unless ( @rv )
		{
			$fnsth->finish() ;
			$fnsth = undef ;
		}
	}

	return @rv ;
}

###############################################################################
#
# transaction stuff

sub SQL_transaction()
{
	$sql_log && print STDERR "SQL: BEGIN TRANSACTION\n" ;
	$dbh->begin_work() ;
}

sub SQL_commit()
{
	$sql_log && print STDERR "SQL: COMMIT\n" ;
	$dbh->commit() ;
}

sub SQL_rollback()
{
	$sql_log && print STDERR "SQL: ROLLBACK\n" ;
	$dbh->rollback() ;
}

sub SQL_true()
{
	if ( $dbconn =~ m|^dbi:sqlite|i )
	{
		return '1' ;
	}

	return 'TRUE' ;
}

sub SQL_false()
{
	if ( $dbconn =~ m|^dbi:sqlite|i )
	{
		return '0' ;
	}

	return 'FALSE' ;
}

###############################################################################
#
# format commas into a number

sub comma($)
{
	my $x = reverse $_[0] ;
	$x =~ s/(\d\d\d)(?=\d)(?!\d*\.)/$1,/g ;
	return scalar reverse $x ;
}

###############################################################################
#
# file and text parsing stuff

sub read_file($)
{
	my $file = shift ;

	my $text = "" ;

	open ( I , "<$file" ) or die "Can\'t read $file: $!\n" ;
	while ( my $line = <I> )
	{
		$text .= $line ;
	}
	close I ;

	return $text ;
}

sub read_template($)
{
	return read_file ( "$docroot$approot/lib/$_[0]" ) ;
}

sub parse_string($;@)
{
	my $string = shift ;
	my %rep = @_ ;

	my @d = localtime ;

	my @accessname =
	(
		"(Not logged in)" ,
		" (Logged in)" ,
		" (Domain Administrator)" ,
		" (System Administrator)" ,
	) ;

	# constants which are always present
	$rep{'g4ver'}     = $g4ver     ;
	$rep{'sitename'}  = $sitename  ;
	$rep{'site_url'}  = $site_url  ;
	$rep{'app_url'}   = $app_url   ;
	$rep{'css_url'}   = $css_url   ;
	$rep{'year'}      = $d[5]+1900 ;
	$rep{'remote_ip'} = $remote_ip ;
	$rep{'userid'}    = ( $userid || "" ) ;
	$rep{'access'}    = $accessname[$access] ;
	$rep{'focus'}     = ( $focus || '' ) ;

	$rep{'logout'}    = $userid
		? "<a href='$app_url/logout.cgi'>[LOGOUT]</a>"
		: '' ;

	# here's where the magic happens
	$string =~ s/\$\{(.+?)\}/$rep{$1}/gms ;

	return $string ;
}

sub parse_file($;@)
{
	my $file = shift ;
	my $text = read_file ( $file ) ;
	return parse_string ( $text , @_ ) ;
}

sub parse_template($;@)
{
	my $file = shift ;
	my $text = read_template ( $file ) ;
	return parse_string ( $text , @_ ) ;
}

###############################################################################
#
# data validation
# these functions return 1 if valid, 0 if not

sub validate_email($)
{
	return ( ( $_[0] =~ m|\@| ) ? 1 : 0 )
}

sub validate_ip($)
{
	if ( $_[0] !~ m|^(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})(/(\d{1,2}))?$| )
	{
		return 0 ;
	}

	my ( $a , $b , $c , $d , $e ) = ( $1 , $2 , $3 , $4 , $6 ) ;

	if ( ( $a > 255 ) || ( $b > 255 ) || ( $c > 255 ) || ( $d > 255 ) )
	{
		return 0 ;
	}

	if ( $e && ( ( $e == 31 ) || ( $e > 32 ) ) )
	{
		return 0 ;
	}

	return 1 ;
}

sub validate_rbl($)
{
	return ( ( $_[0] =~ m|.+| ) ? 1 : 0 ) ;
}

sub validate_grey($)
{
	return ( ( $_[0] =~ m|^\d+$| ) ? 1 : 0 ) ;
}

sub validate_regex($)
{
	my $pat = shift ;
	my $re = eval { no strict ; no warnings ; qr|$pat| } ;
	return ( $@ ? 0 : 1 ) ;
}


###############################################################################
#
# web stuff

sub cookies()
{
	my $c = '' ;
	my $sec = ( ( $web_proto eq "https" ) ? " secure=1;" : "" ) ;

	if ( $session )
	{
		$c .= "Set-cookie: session=$session; path=$approot;$sec\n" ;

		if ( $access > 1 )
		{
			$c .= "Set-cookie: focus=$focus; path=$approot;$sec\n" ;
		}
	}
	else
	{
		########################################
		# if no session, then explicitly delete any values in browser

		if ( $rc_session )
		{
			$c .= "Set-cookie: session=x; path=$approot;$sec"
				. " expires=Thu, 01 Jan 1970 00:00:00 GMT;\n" ;
		}

		if ( $rc_focus )
		{
			$c .= "Set-cookie: focus=x; path=$approot;$sec"
				. " expires=Thu, 01 Jan 1970 00:00:00 GMT;\n" ;
		}
	}

	$c =~ s|\n$||s ;
	return $c ;
}

sub redirect_url($)
{
	my $url = shift ;
	if ( $url !~ m|://| )
	{
		$url = $site_url . $url ;
	}

	my $cookies = cookies() ;

	print <<EOF ;
Status: 303
Content-type: text/html; charset=UTF-8
Location: $url
Connection: close
Cache-control: no-store
Expires: Thu, 01 Jan 1970 00:00:00 GMT
$cookies

<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
   "http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<title>Redirect</title>
</head>
<body>
<p>Please <a href="$url">click here</a> if your browser does not
automatically redirect you.</p>
</body>
</html>
EOF
	exit 0 ;
}

sub redirect_self()
{
	my $request_uri = ( $ENV{"REQUEST_URI"} || "" ) ;

	fatal ( "redirect_self: no REQUEST_URI found" )
		unless ( $request_uri ) ;

	my $url = $site_url ;
	if ( $request_uri =~ m|^/| )
	{
		$url =~ s|/$|| ;
	}

	redirect_url ( $url . $request_uri ) ;
}

sub redirect_home()
{
	redirect_url ( $app_url . '/' ) ;
}

###############################################################################
###############################################################################
###############################################################################
#
# Web session handling code
#
# get_session: look up a session ID in the database.
# If found, return the IP, userid, and level.
# Otherwise, return undef.

sub get_session($)
{
	my $sid = shift ;
	my $qsid = SQL_quote ( $sid ) ;
	my $now = time() ;

	my $query = "SELECT ip , userid , level FROM sessions"
		. " WHERE id = $qsid"
		. " AND expires > $now"
		. " AND NOT ended" ;
	my ( $ip , $userid , $level ) = SQL_select ( $query ) ;

	unless ( $ip && $userid && $level )
	{
		return undef ;
	}

	$ip =~ s|\s+$|| ;
	return ( $ip , $userid , $level ) ;
}

###############################################################################
#
# Create a new session.

sub create_session()
{
	unless ( $userid && $access )
	{
		fatal "create_session() without userid/access\n" ;
	}

	unless ( $session )
	{
		$session = mk_sid() ;
		my $now = time() ;
		my $exp = $now + $ses_timeout ;
	
		my $query = "INSERT INTO sessions ( id , ip , userid , level"
			. " , first , expires ) VALUES ( "
			. SQL_quote ( $session   ) . ' , '
			. SQL_quote ( $remote_ip ) . ' , '
			. SQL_quote ( $userid    ) . ' , '
			. " $access , $now , $exp )" ;
		SQL_do ( $query ) ;
	}
}

###############################################################################
#
# End the current session, or invalidate a specific session key.

sub end_session(;$)
{
	my $s = ( shift || $session ) ;
	return unless ( $s ) ;

	my $query = "UPDATE sessions SET ended = " . SQL_true()
		. " WHERE id = " . SQL_quote ( $s ) ;
	SQL_do ( $query ) ;

	if ( $s eq $session )
	{
		$session = '' ;
		$userid = '' ;
		$access = 0 ;
		$focus = '' ;
	}
}

###############################################################################
#
# mk_sid: Create a session ID which isn't already being used.

sub mk_sid()
{
	my $t = time() ;
	my $s = 0 ;
	while ( ++$s )
	{
		my $raw = sprintf ( "%d.%d.%d.skey" , $t , $$ , $s ) ;
		my $md5 = md5_base64 ( $raw ) ;

		# Any "+" in the ID causes issues with the browsers mangling
		# them, something to do with URL encoding. Convert any "+"
		# characters to "-" instead.
		$md5 =~ s|\+|\-|g ;

		my ( $ip , $userid , $level ) = get_session ( $md5 ) ;
		unless ( $ip || $userid || $level )
		{
			return $md5 ;
		}
	}
}

###############################################################################
#
# set_focus: Set the $focus variable to the indicated value, if valid for
# the current access level. If not, set it to $userid.

sub set_focus($)
{
	my $try = shift ;

	########################################
	# Handle focus cookie.
	# $focus will end up as one of the following:
	# - For mailbox access, only $userid.
	# - For domain access, must end with "@domain" of $userid.
	# - For system access, must contain "@".

	if ( 3 == $access )	# system admin access
	{
		$focus = ( ( $try =~ m|\@| ) ? $try : $userid ) ;
	}
	elsif ( 2 == $access )	# domain admin access
	{
		( my $udom = $userid ) =~ s|^.*\@|| ;
		$focus = ( ( $try =~ m|\@$udom$| ) ? $try : $userid ) ;
	}
	else			# mailbox-only access
	{
		$focus = $userid ;
	}

	return $focus ;
}

###############################################################################
#
# turn a $focus value into a consistent string

sub focus_label(;$)
{
	my $pf = ( shift || $focus ) ;

	if ( $pf eq '@' )
	{
		return 'Server' ;
	}

	if ( $pf =~ m|^\@| )
	{
		return "Domain $pf" ;
	}

	return $pf ;
}

sub focus_string(;$)
{
	my $pf = ( shift || $focus ) ;

	if ( $pf eq '@' )
	{
		return 'the entire server' ;
	}

	if ( $pf =~ m|^\@| )
	{
		return "the domain $pf" ;
	}

	return $pf ;
}

###############################################################################
#
# get a list of all entities which currently have rules in the database
# optionally filter the list to a single domain
#
# returns an HTML list with links to focus.cgi

sub get_focus_list(;$)
{
	my $pdom = ( shift || '' ) ;

	########################################
	# get the raw list form the database

	my $query = 'SELECT DISTINCT recipient FROM rules' ;
	if ( $pdom )
	{
		$query .= ' WHERE recipient LIKE '
			. SQL_quote ( "%\@$pdom" ) ;
	}
	$query .= ' ORDER BY recipient' ;

	my @recip = SQL_select ( $query ) ;

	########################################
	# correlate by domain

	my %domain = () ;	# domain => user@domain,user@domain,...

	for my $r ( @recip )
	{
		$r =~ m|^(.*?)\@(.*)$| ;
		my ( $user , $dom ) = ( $1 , $2 ) ;
		next unless ( $dom ) ;

		if ( $user eq '%' )
		{
			$domain{$dom} ||= '' ;
			next ;
		}

		$domain{$dom} .= ( $domain{$dom} ? ',' : '' ) ;
		$domain{$dom} .= $r ;
	}

	########################################
	# build output list

	my $output = "<ul>\n" ;

	unless ( $pdom )
	{
		$output .= "<li><a href='focus.cgi?type=s'>System-wide rules</a></li>\n" ;
	}

	for my $d ( sort keys %domain )
	{
		my $zdom = ( ( $access > 2 ) ? "&dom=$d" : '' ) ;
		$output .= "<li><a href='focus.cgi?type=d$zdom'>Domain <tt>$d</tt></a></li>\n" ;
		if ( $domain{$d} ne '' )
		{
			$output .= "<ul>\n" ;
			for my $r ( split ( m|,| , $domain{$d} ) )
			{
				if ( $access > 2 )
				{
					$output .= "<li><a href='focus.cgi?type=m&mbox=$r'><tt>$r</tt></a></li>\n" ;
				}
				elsif ( $access > 1 )
				{
					$r =~ m|^(.*)\@| ;
					my $u = $1 ;
					$output .= "<li><a href='focus.cgi?type=u&user=$u'><tt>$r</tt></a></li>\n" ;
				}
			}
			$output .= "</ul>\n" ;
		}
	}

	$output .= "</ul>\n" ;

	return $output ;
}

###############################################################################
#
# rule_is_final: return 1 for "final" rule types, 0 otherwise.

my %rule_final = qw
(
	A	1
	D	0
	E	0
	G	1
	I	0
	R	0
	T	0
	U	0
) ;

sub rule_is_final($)
{
	return ( $rule_final{$_[0]} || 0 ) ;
}

###############################################################################
#
# rule_can_edit: return 1 if rule type allows for editing, 0 otherwise

my %rule_editable = qw
(
	A	1
	D	1
	E	1
	G	1
	I	1
	R	1
	T	1
	U	1
) ;

sub rule_can_edit($)
{
	return ( $rule_editable{$_[0]} || 0 ) ;
}

###############################################################################
#
# init: Initialize a new request. If a session cookie is present, decode it.

my $init_done = 0 ;

sub init()
{
	########################################
	# don't waste CPU cycles
	# (no need to do vpopmaild check multiple times)

	$init_done && return ;
	$init_done = 1 ;

	########################################
	# Configure the database connection.

	$dbconn ||= ( $ENV{'G4DBCONN'} || '' ) ;
	$dbuser ||= ( $ENV{'G4DBUSER'} || '' ) ;
	$dbpass ||= ( $ENV{'G4DBPASS'} || '' ) ;

	########################################
	# Get CGI data and cookie values.

	$docroot	= ( $ENV{"DOCUMENT_ROOT"} || $docroot ) ;
	$request_uri	= ( $ENV{"REQUEST_URI"}   || "" ) ;
	$remote_ip	= ( $ENV{"REMOTE_ADDR"}   || "" ) ;

	$web_proto	= $ENV{"SSL_CIPHER"} ? "https" : "http" ;
	my $sname	= ( $ENV{"SERVER_NAME"} || "" ) ;
	my $sport	= ( $ENV{"SERVER_PORT"} || "" ) ;

	$approot	= $request_uri ;
	$approot	=~ s|/[^/]*$|| ;
	$css_url	= "$approot/rules.css" ;

	$rc_session	= ( cookie ( 'session' ) || '' ) ;
	$rc_focus	= ( cookie ( 'focus'   ) || '' ) ;

	if ( $sname )
	{
		$site_url = "$web_proto://$sname" ;
		if ( $sport != ( ( $web_proto eq "https" ) ? 443 : 80 ) )
		{
			$site_url .= ":$sport" ;
		}
		$app_url = $site_url . $approot ;
	}

	########################################
	# Handle session cookie, if present.

	$session = '' ;
	$userid = '' ;
	$access = 0 ;
	$focus = '' ;

	if ( $rc_session )
	{
		my ( $s_ip , $s_user , $s_level )
			= get_session ( $rc_session ) ;
		if ( $s_ip )
		{
			if ( $s_ip eq $remote_ip )
			{
				$userid  = $s_user ;
				$access  = $s_level ;
				$session = $rc_session ;
			}
			else	# session is on the wrong IP
			{
				logit ( "Ending session \"$rc_session\","
					. " wrong IP. session=[$s_ip]"
					. " client=[$remote_ip]\n" ) ;
				end_session ( $rc_session ) ;
			}
		}
	}

	########################################
	# If we have a valid session, update the session record
	# and read the focus cookie.

	if ( $session )
	{
		fatal 'invalid SESSION_TIMEOUT value'
			unless ( $ses_timeout =~ m|^\d+$| ) ;

		my $exp = time() + $ses_timeout ;

		my $q = 'UPDATE sessions'
			. " SET expires = $exp , hits = hits + 1"
			. ' WHERE id = ' . SQL_quote ( $session ) ;
		SQL_do ( $q ) ;

		########################################
		# Handle focus cookie.

		set_focus ( $rc_focus || $userid ) ;
	}

	########################################
	# Read $g4ver from first line of g4ver.txt

	$g4ver = read_template ( 'g4ver.txt' ) ;
	$g4ver =~ s|\n.*||s ;
	$g4ver =~ s|\s+$||s ;
}

###############################################################################
#
# page start/end

sub page_start(;$)
{
	my $title = ( shift || "" ) ;

	my $ztitle	= ( $title ? "$sitename - $title" : $sitename ) ;
	my $zh1		= ( $title ? $title : $sitename ) ;
	my $cookies	= cookies() ;
	my $modeline	= "<!-- access=$access -->" ;
	my $uri		= $request_uri ;

#	if ( $access > 1 )
#	{
#		my $fl = focus_label() ;
#		$modeline .= "<a href='$app_url/focus.cgi'>[FOCUS ($fl)]</a> " ;
#	}

	return parse_template ( 'page-start.txt' ,
		'title'		=> $ztitle ,
		'h1'		=> $zh1 ,
		'cookies'	=> $cookies ,
		'modeline'	=> $modeline ) ;
}

sub page_end()
{
	return parse_template ( "page-end.txt" ) ;
}

###############################################################################
#
# check ID/password using vpopmaild
#
# returns access level:
#	0 not logged in
#	1 normal user, mailbox access
#	2 domain admin
#	3 system admin
#
# also sets global $login_err

sub check_vpopmaild($$)
{
	my $id = shift ;
	my $pw = shift ;

	my $ok = 0 ;
	my $da = 0 ;
	my $sa = 0 ;

	$login_err = "" ;

	########################################
	# Identify the vpopmaild server we will be connecting to

	my $vpd_host = ( $ENV{'VPOPMAILD_HOST'} || '127.0.0.1' ) ;
	my $vpd_port = ( $ENV{'VPOPMAILD_PORT'} || 89 ) ;

	########################################
	# create socket

	my $paddr = sockaddr_in ( $vpd_port , inet_aton ( $vpd_host ) ) ;
	my $proto = getprotobyname ( "tcp" ) ;

	unless ( socket ( S , PF_INET , SOCK_STREAM , $proto ) )
	{
		$login_err = "Can't create socket: $!\n" ;
		return 0 ;
	}

	########################################
	# set socket for non-buffered output
	# also set for binary mode (???)

	my $oh = select ( S ) ;
	$| = 1 ;
	select $oh ;

	binmode S ;

	########################################
	# connect to vpopmaild

	unless ( connect ( S , $paddr ) )
	{
		$login_err = "Can't connect to vpopmaild: $!\n" ;
		return 0 ;
	}

	########################################
	# banner should be "+OK\r\n"

	my $resp = <S> ;
	unless ( $resp =~ /^\+/ )
	{
		$login_err = "Invalid banner from vpopmaild" ;
		close S ;
		return 0 ;
	}

	########################################
	# send login request

	print S "login $id $pw\n" ;

	########################################
	# response will be "+OK+" and multiple lines,
	# or "-ERR"

	$resp = <S> ;
	$resp =~ s/\s+$// ;

	unless ( $resp =~ /^\+/ )
	{
		$login_err = $resp ;
		close S ;

		print STDERR "<$id> $resp" ;

		return 0 ;
	}

	########################################
	# we know the credentials are good.
	# do they have user, domain, or system level access?

	while ( $resp = <S> )
	{
		last if ( $resp =~ /^\./ ) ;
		if ( $resp =~ /^domain_admin_privileges (\d+)/ )
		{
			$da = $1 ;
		}

		if ( $resp =~ /^system_admin_privileges (\d+)/ )
		{
			$sa = $1 ;
		}
	}

	########################################
	# buh-bye now

	print S "quit\n" ;
	$resp = <S> ;
	close S ;

	########################################
	# and i'm outta here as well

	return ( $sa ? 3 : ( $da ? 2 : 1 ) ) ;
}

###############################################################################
#
# generic function to require login

sub require_login()
{
	init() ;

	$access || redirect_url ( "$app_url/login.cgi" ) ;
}

###############################################################################
#
# make sure we can run

1 ;
